@@ -1444,6 +1444,10 @@ class LineageRecord(PBObject):
                                                                              self.pb.code_hash)
         return s
 
+    @staticmethod
+    def dependency_count(lineage_pb):
+        return len(lineage_pb.depends_on)
+
     @staticmethod
     def add_deps_to_lr(lineage_pb, depends_on):
         _ = [lineage_pb.depends_on.add(hframe_proc_name=tup[0],
@@ -1779,7 +1783,7 @@ class FrameRecord(PBObject):
             'big': hyperframe_pb2.BIG,
         }
 
-        if numpy_byteorder is '=':
+        if numpy_byteorder == '=':
             return system_endianness[sys.byteorder]
 
         return numpy_endianness[numpy_byteorder]
